@@ -33,17 +33,23 @@ import glob
 import traceback
 import atexit
 import signal
+import time
 
 import rclexecm
 import rclconfig
 
 _mswindows = (sys.platform == "win32")
 
-# Can we access the poppler-glib python3 bindings ? This would allow extracting
-# text from annotations.
+# Test access to the poppler-glib python3 bindings ? This allows
+# extracting text from annotations.
 # - On Ubuntu, this comes with package gir1.2-poppler-0.18
 # - On opensuse the package is named typelib-1_0-Poppler-0_18
 # (actual versions may differ of course).
+# 
+# NOTE: we are using the glib introspection bindings for the Poppler C
+# API, which is not the same as the Python bindings of the Poppler C++
+# API (poppler-cpp). The interface is quite different
+# (e.g. attachments are named "embeddedfiles" in the C++ interface.
 havepopplerglib = False
 try:
     import gi
@@ -61,11 +67,17 @@ _htmlprefix =b'''<html><head>
 _htmlsuffix = b'''</pre></body></html>'''
 
 def finalcleanup():
+    global tmpdir
     if tmpdir:
-        vacuumdir(tmpdir)
-        os.rmdir(tmpdir)
+        del tmpdir
+        tmpdir = None
 
+ocrproc = None
 def signal_handler(signal, frame):
+    global ocrproc
+    if ocrproc:
+        ocrproc.wait()
+        ocrproc = None
     sys.exit(1)
 
 atexit.register(finalcleanup)
@@ -80,22 +92,21 @@ except: pass
 try: signal.signal(signal.SIGTERM, signal_handler)
 except: pass
 
-def vacuumdir(dir):
-    if dir:
-        for fn in os.listdir(dir):
-            path = os.path.join(dir, fn)
-            if os.path.isfile(path):
-                os.unlink(path)
-    return True
-
 class PDFExtractor:
     def __init__(self, em):
         self.currentindex = 0
         self.pdftotext = None
+        self.pdftotextversion = 0
         self.pdfinfo = None
+        self.pdfinfoversion = 0
         self.pdftk = None
         self.em = em
         self.tesseract = None
+        self.extrameta = None
+        self.re_head = None
+
+        self.config = rclconfig.RclConfig()
+        self.confdir = self.config.getConfDir()
 
         # Avoid picking up a default version on Windows, we want ours
         if not _mswindows:
@@ -106,31 +117,39 @@ class PDFExtractor:
                 # No need for anything else. openfile() will return an
                 # error at once
                 return
+        self.pdftotextversion = self._popplerutilversion(self.pdftotext)
+        # Check if we need to escape portions of text: old versions of pdftotext output raw
+        # HTML special characters. Don't know exactly when this changed but it's fixed in 0.26.5
+        if self.pdftotextversion >= 2605:
+            self.needescape = False
+        else:
+            self.needescape = True
 
-        self.config = rclconfig.RclConfig()
-        self.confdir = self.config.getConfDir()
-        # The user can set a list of meta tags to be extracted from
-        # the XMP metadata packet. These are specified as
-        # (xmltag,rcltag) pairs
-        self.extrameta = self.config.getConfParam("pdfextrameta")
-        if self.extrameta:
-            self.extrametafix = self.config.getConfParam("pdfextrametafix")
-            self._initextrameta()
+        # pdfinfo may be used to extract XML metadata and custom PDF properties
+        if not _mswindows:
+            self.pdfinfo = rclexecm.which("pdfinfo")
+        if not self.pdfinfo:
+            self.pdfinfo = rclexecm.which("poppler/pdfinfo")
+        if self.pdfinfo:
+            self.pdfinfoversion = self._popplerutilversion(self.pdfinfo)
+            # The user can set a list of meta tags to be extracted from the XMP metadata
+            # packet. These are specified as (xmltag,rcltag) pairs
+            self.extrameta = self.config.getConfParam("pdfextrameta")
+            if self.extrameta:
+                self.extrametafix = self.config.getConfParam("pdfextrametafix")
+                self._initextrameta()
+        #self.em.rclog(f"PDFINFOVERSION {self.pdfinfoversion}")
 
-        # Check if we need to escape portions of text where old
-        # versions of pdftotext output raw HTML special characters.
-        self.needescape = True
-        try:
-            version = subprocess.check_output([self.pdftotext, "-v"],
-                                              stderr=subprocess.STDOUT)
-            major,minor,rev = version.split()[2].split('.')
-            # Don't know exactly when this changed but it's fixed in
-            # jessie 0.26.5
-            if int(major) > 0 or int(minor) >= 26:
-                self.needescape = False
-        except:
-            pass
-        
+        # Extracting the outline / bookmarks needs still another poppler command...
+        self.dooutline = self.config.getConfParam("pdfoutline")
+        if self.dooutline:
+            if not _mswindows:
+                self.pdftohtml = rclexecm.which("pdftohtml")
+            if not self.pdftohtml:
+                self.pdftohtml = rclexecm.which("poppler/pdftohtml")
+            if not self.pdftohtml:
+                self.dooutline = False
+            
         # Pdftk is optionally used to extract attachments. This takes
         # a hit on performance even in the absence of any attachments,
         # so it can be disabled in the configuration.
@@ -143,15 +162,21 @@ class PDFExtractor:
         if self.pdftk:
             self.maybemaketmpdir()
 
+    def _popplerutilversion(self, tool):
+        try:
+            output = subprocess.check_output([tool, "-v"], stderr=subprocess.STDOUT)
+            l1 = output.split(b"\n")[0]
+            v = l1.split()[2]
+            M,m,r = v.split(b".")
+            version = int(M) * 10000 + int(m) * 100 + int(r)
+        except:
+            version = 0
+        return version
+    
     def _initextrameta(self):
-        if not _mswindows:
-            self.pdfinfo = rclexecm.which("pdfinfo")
         if not self.pdfinfo:
-            self.pdfinfo = rclexecm.which("poppler/pdfinfo")
-        if not self.pdfinfo:
-            self.extrameta = None
             return
-
+        
         # extrameta is like "metanm|rclnm ...", where |rclnm maybe absent (keep
         # original name). Parse into a list of pairs.
         l = self.extrameta.split()
@@ -179,7 +204,6 @@ class PDFExtractor:
             self.pdfinfo = None
             return
 
-        self.re_head = re.compile(br'<head>', re.IGNORECASE)
         self.re_xmlpacket = re.compile(br'<\?xpacket[ 	]+begin.*\?>' +
                                        br'(.*)' + br'<\?xpacket[ 	]+end',
                                        flags = re.DOTALL)
@@ -197,6 +221,7 @@ class PDFExtractor:
                 EMF = None
                 pass
 
+
     # Extract all attachments if any into temporary directory
     def extractAttach(self):
         if self.attextractdone:
@@ -208,25 +233,48 @@ class PDFExtractor:
             # no big deal
             return True
         try:
-            vacuumdir(tmpdir)
+            tmpdir.vacuumdir()
             # Note: the java version of pdftk sometimes/often fails
             # here with writing to stdout:
             #    Error occurred during initialization of VM
             #    Could not allocate metaspace: 1073741824 bytes
-            # Maybe unsufficient resources when started from Python ?
+            # Maybe insufficient resources when started from Python ?
             # In any case, the important thing is to discard the
             # output, until we fix the error or preferably find a way
             # to do it with poppler...
             subprocess.check_call(
-                [self.pdftk, self.filename, "unpack_files", "output",
-                 tmpdir], stdout=sys.stderr)
-            self.attachlist = sorted(os.listdir(tmpdir))
+                [self.pdftk, self.filename, "unpack_files", "output", tmpdir.getpath()],
+                stdout=sys.stderr)
+            self.attachlist = sorted(os.listdir(tmpdir.getpath()))
             return True
         except Exception as e:
             self.em.rclog("extractAttach: failed: %s" % e)
             # Return true anyway, pdf attachments are no big deal
             return True
 
+
+    # Extract PDF custom properties into a dict. Only works with newer pdfinfo versions
+    def _customfields(self):
+        if not self.pdfinfo or self.pdfinfoversion < 211000:
+            return {}
+        try:
+            fields = {}
+            output = subprocess.check_output([self.pdfinfo, "-custom", self.filename],
+                                             stderr=subprocess.STDOUT)
+            output = output.decode("utf-8")
+            for line in output.split("\n"):
+                colon = line.find(":")
+                if colon > 0:
+                    fld = line[:colon].strip()
+                    value = line[colon+1:].strip()
+                    if fld not in ("Author", "CreationDate", "Creator", "ModDate",
+                                   "Producer", "Subject", "Title"):
+                        fields[fld] = value
+            return fields
+        except:
+            return {}
+        
+
     # pdftotext (used to?) badly escape text inside the header
     # fields. We do it here. This is not an html parser, and depends a
     # lot on the actual format output by pdftotext.
@@ -238,6 +286,9 @@ class PDFExtractor:
         didcs = False
         output = []
         isempty = True
+        fields = {}
+        if self.pdfinfoversion > 211000:
+            fields = self._customfields()
         for line in input.split(b'\n'):
             if re.search(b'</head>', line):
                 inheader = False
@@ -247,6 +298,8 @@ class PDFExtractor:
                 if not didcs:
                     output.append(b'<meta http-equiv="Content-Type"' + \
                               b'content="text/html; charset=UTF-8">\n')
+                    for fld,val in fields.items():
+                        output.append(self._metatag(fld, val))
                     didcs = True
                 if self.needescape:
                     m = re.search(b'''(.*<title>)(.*)(<\/title>.*)''', line)
@@ -279,12 +332,16 @@ class PDFExtractor:
 
         return b'\n'.join(output), isempty
 
+
     def _metatag(self, nm, val):
         return b"<meta name=\"" + rclexecm.makebytes(nm) + b"\" content=\"" + \
                rclexecm.htmlescape(rclexecm.makebytes(val)) + b"\">"
 
+
     # metaheaders is a list of (nm, value) pairs
     def _injectmeta(self, html, metaheaders):
+        if self.re_head is None:
+            self.re_head = re.compile(br'<head>', re.IGNORECASE)
         metatxt = b''
         for nm, val in metaheaders:
             metatxt += self._metatag(nm, val) + b'\n'
@@ -317,7 +374,7 @@ class PDFExtractor:
         all = subprocess.check_output([self.pdfinfo, "-meta", self.filename])
         res = self.re_xmlpacket.search(all)
         xml = res.group(1) if res else ''
-        # self.em.rclog("extrameta: XML: [%s]" % xml)
+        #self.em.rclog(f"extrameta: XML: [{xml}]")
         if not xml:
             return html
 
@@ -326,14 +383,21 @@ class PDFExtractor:
         # the stackoverflow ref above. Maybe we'd be better off just
         # walking the full tree and building the namespaces dict.
         root = ET.fromstring(xml)
-        #self.em.rclog("NSMAP: %s"% root.nsmap)
+        #self.em.rclog(f"root nsmap: {root.nsmap}")
         namespaces = {'rdf' : "http://www.w3.org/1999/02/22-rdf-syntax-ns#"}
-        rdf = root.find("rdf:RDF", namespaces)
-        #self.em.rclog("RDF NSMAP: %s"% rdf.nsmap)
+        RDFROOTTAG = "{" + namespaces["rdf"] + "}RDF"
+
+        if root.tag == RDFROOTTAG:
+            rdf = root
+        else:
+            rdf = root.find("rdf:RDF", namespaces)
         if rdf is None:
+            #self.em.rclog(f"_setextrameta: rdf is none")
             return html
+        #self.em.rclog("rdf nsmap: %s"% rdf.nsmap)
         rdfdesclist = rdf.findall("rdf:Description", rdf.nsmap)
         for metanm,rclnm in self.extrameta:
+            #self.em.rclog(f"_setextrameta: metanm {metanm} rclnm {rclnm}")
             for rdfdesc in rdfdesclist:
                 try:
                     elts = rdfdesc.findall(metanm, rdfdesc.nsmap)
@@ -382,6 +446,7 @@ class PDFExtractor:
                             pass
                         metaheaders.append((rclnm, text))
         if metaheaders:
+            #self.em.rclog(f"extrameta: {metaheaders}")
             if emf:
                 try:
                     emf.wrapup(metaheaders)
@@ -394,11 +459,12 @@ class PDFExtractor:
     def maybemaketmpdir(self):
         global tmpdir
         if tmpdir:
-            if not vacuumdir(tmpdir):
-                self.em.rclog("openfile: vacuumdir %s failed" % tmpdir)
+            if not tmpdir.vacuumdir():
+                self.em.rclog("openfile: vacuumdir %s failed" % tmpdir.getpath())
                 return False
         else:
-            tmpdir = tempfile.mkdtemp(prefix='rclmpdf')
+            tmpdir = rclexecm.SafeTmpDir("rclpdf", self.em)
+            #self.em.rclog("Using temporary directory %s" % tmpdir.getpath())
             if self.pdftk and re.match("/snap/", self.pdftk):
                 # We know this is Unix (Ubuntu actually). Check that tmpdir
                 # belongs to the user as snap commands can't use /tmp to share
@@ -410,9 +476,7 @@ class PDFExtractor:
                     if st.st_uid == os.getuid():
                         ok = True
                 if not ok:
-                    self.em.rclog(
-                        "pdftk is a snap command and needs TMPDIR to be "
-                        "a directory you own")
+                    self.em.rclog("pdftk is a snap command and needs TMPDIR to be owned by you")
 
     def _process_annotations(self, html):
         doc = Poppler.Document.new_from_file(
@@ -464,7 +528,39 @@ class PDFExtractor:
         if annotsfield:
             self.em.setfield("pdfannot", annotsfield)
         return html
-    
+
+
+    def _process_outline(self):
+        import xml.sax
+        data = subprocess.check_output([self.pdftohtml, "-i", "-s", "-enc", "UTF-8", "-xml",
+                                 "-stdout", self.filename])
+        class OutlineHandler(xml.sax.ContentHandler):
+            def __init__(self):
+                self.outlinelevel = 0
+                self.page = 0
+                self.alltext = ""
+                self.initem = False
+            def startElement(self, name, attrs):
+                if name == "outline":
+                    self.outlinelevel += 1
+                elif name == "item" and self.outlinelevel:
+                    self.initem = True
+                    if "page" in attrs:
+                        self.alltext += f"[P. {attrs['page']}] "
+            def endElement(self, name):
+                if name == "item" and self.outlinelevel:
+                    self.initem = False
+                    self.alltext += "\n"
+                elif name == "outline":
+                    self.outlinelevel -= 1
+            def characters(self, content):
+                if self.initem:
+                    self.alltext += content
+        handler = OutlineHandler()
+        xml.sax.parseString(data, handler)
+        return handler.alltext
+
+        
     def _selfdoc(self):
         '''Extract the text from the pdf doc (as opposed to attachment)'''
         self.em.setmimetype('text/html')
@@ -486,27 +582,36 @@ class PDFExtractor:
             s = self.config.getConfParam("pdfocr")
             if rclexecm.configparamtrue(s):
                 try:
-                    cmd = [sys.executable, os.path.join(_execdir, "rclocr.py"),
-                           self.filename]
-                    data = subprocess.check_output(cmd)
+                    cmd = [sys.executable, os.path.join(_execdir, "rclocr.py"), self.filename]
+                    global ocrproc
+                    ocrproc = subprocess.Popen(cmd, stdout=subprocess.PIPE)
+                    data, stderr = ocrproc.communicate()
+                    ocrproc = None
                     html = _htmlprefix + rclexecm.htmlescape(data) + _htmlsuffix
                 except Exception as e:
-                    self.em.rclog("%s failed: %s" % (cmd, e))
+                    self.em.rclog(f"{cmd} failed: {e}")
                     pass
 
         if self.extrameta:
             try:
                 html = self._setextrameta(html)
             except Exception as err:
-                self.em.rclog("Metadata extraction failed: %s %s" %
-                              (err, traceback.format_exc()))
+                self.em.rclog(f"Metadata extraction failed: {err} {traceback.format_exc()}")
+
+        if self.dooutline:
+            try:
+                outlinetext = self._process_outline()
+                html = self._injectmeta(html, [("description" , outlinetext),])
+            except Exception as err:
+                self.em.rclog(f"Outline extraction failed: {err} {traceback.format_exc()}")
 
         if havepopplerglib:
             try:
                 html = self._process_annotations(html)
             except Exception as err:
-                self.em.rclog("Annotation extraction failed: %s %s" %
-                              (err, traceback.format_exc()))
+                self.em.rclog(f"Annotation extraction failed: {err} {traceback.format_exc()}")
+
+            
         return (True, html, "", eof)
 
 
@@ -515,7 +620,9 @@ class PDFExtractor:
         if not self.attextractdone:
             if not self.extractAttach():
                 return (False, "", "", rclexecm.RclExecM.eofnow)
-        path = os.path.join(tmpdir, ipath)
+        if type(ipath) != type(""):
+            ipath = ipath.decode('utf-8')
+        path = os.path.join(tmpdir.getpath(), ipath)
         if os.path.isfile(path):
             f = open(path, "rb")
             docdata = f.read();
